##############################################################################
#
# Copyright (c) 2003 gocept gmbh & co. kg. All rights reserved.
#
# See also LICENSE.txt
#
##############################################################################
"""CMF link checker tool
$Id: LinkCheckerTool.py,v 1.6 2003/12/17 19:18:27 ctheune Exp $"""

# Python stuff
from types import StringTypes
from md5 import md5
from random import shuffle

# Global Zope stuff
from DateTime import DateTime
from AccessControl import ClassSecurityInfo
from OFS.Folder import Folder
from Globals import InitializeClass, PersistentMapping, Persistent
from Products.CMFCore.utils import UniqueObject
from Products.CMFCore.CMFCorePermissions import ManagePortal
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.CMFCore.ActionProviderBase import ActionProviderBase
from Products.CMFCore.ActionInformation import ActionInformation
from Products.CMFCore.Expression import Expression
from AccessControl import getSecurityManager

# Sibling imports
from interfaces import ILinkCheckerTool, ILinkInformation
from LinkInfo import LinkInfo
import retrievers

import zLOG

# Threading stuff
from threading import Thread
try:
    from threading import BoundedSemaphore
except ImportError:
    from threading import Semaphore
    BoundedSemaphore = Semaphore
import urllib, httplib
import httpclient
import Queue
import time

# Threading and transactions are a pain
import Globals
import OFS.Application
import ZODB.ZApplication

# Link checker stuff
from httplib import ResponseNotReady
from httpcodes import http_codes

retrieval_queue = Queue.Queue(0) # Make this an infinite synchronized queue
thread_stop = 0

active_threads = 5
thread_count = 0
pool_threads = BoundedSemaphore(value=active_threads)

class TooManyThreadsRunning(Exception):
    pass

class UnknownLink(Exception):
    pass

class LinkCheckerTool(UniqueObject, ActionProviderBase, Folder):
    """A link checker tool. Encapsulating mechanisms for finding links in the
       site, testing their availability and notifying their owners."""

    __implements__ = (ILinkCheckerTool, )

    id = 'portal_linkchecker'
    meta_type = 'CMF Linkchecker Tool'
    security = ClassSecurityInfo()
    plone_tool = 1

    manage_options = ( ({'label' : 'Overview', 'action' : 'manage_overview'},) +
                        ActionProviderBase.manage_options +
                       ({'label' : 'Configure', 'action' : 'manage_configure'},
                        {'label' : 'Linklist', 'action' : 'manage_checklinks'},
                        {'label' : 'Linkchecking', 'action' : 'manage_checker'},) +
                        Folder.manage_options )

    _actions = [ActionInformation(id='linkchecker_object_status',
                                  title='Links',
                                  action=Expression(text='string:${object_url}/lc_object_status'),
                                  category='object',
                                  visible=1),
                ActionInformation(id='linkchecker_member_overview',
                                 title='My Links',
                                 action=Expression(text='string:${portal_url}/lc_my_dead_links'),
                                 category='user',
                                 visible=1),
                ActionInformation(id='linkchecker_balanced_scorecard',
                                 title='Balanced Score Card: Links',
                                 category='portal_tabs',
                                 visible=1,
                                 action=Expression(text='string:${portal_url}/lc_balanced_score_card'))]

    maximumCacheAge = 60*60     # 1 hour
    gracefulUpdateAge = 60*45   # 45 minutes
    checkingPeriod = 60*5       # 5 minutes
    defaultURIPrefix = ""       # No prefix by default

    #
    # ZMI methods
    #
    security.declareProtected(ManagePortal, 'manage_overview')
    manage_overview = PageTemplateFile('www/explainLinkcheckerTool', globals(),
                                        __name__='manage_overview')

    security.declareProtected(ManagePortal, 'manage_configure')
    manage_configure= PageTemplateFile('www/configureLinkcheckerTool', globals(),
                                        __name__='manage_configure')

    security.declareProtected(ManagePortal, 'manage_checklinks')
    manage_checklinks = PageTemplateFile('www/checkLinks', globals(),
                                        __name__='manage_checklinks')
         
    security.declareProtected(ManagePortal, 'manage_checker')
    manage_checker = PageTemplateFile('www/checker', globals(),
                                        __name__='manage_checker')
                                        
    security.declareProtected(ManagePortal, 'edit_configuration')
    def edit_configuration(self, maxage, graceage, defaultURIPrefix):
        """Update the configuration options."""
        self.setMaximumCacheAge(maxage)
        self.setGracefulUpdateAge(graceage)
        self.setDefaultURIPrefix(defaultURIPrefix)
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_configure')

    security.declareProtected(ManagePortal, 'manage_checkLink')
    def manage_checkLink(self, id):
        """ZMI method for checking a link."""
        if not self._cache.has_key(url):
            if hasattr(self.REQUEST, 'RESPONSE'):
                self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=No+such+link')
                       
        link = getattr(self, id)
        self.checkLink(link)
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=Link+checked')

    def manage_forceCheckLink(self, url):
        """ZMI method for forcing the check of a link."""
        if not self._cache.has_key(url):
            if hasattr(self.REQUEST, 'RESPONSE'):
                self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=No+such+link')
            return
        link = getattr(self, id)
        self.__queueAddLink(link)
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=Check+forced')

    security.declareProtected(ManagePortal, 'manage_checkAll')
    def manage_checkAll(self):
        """ZMI method for checking all links."""
        self.checkAllLinks()
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_checker?manage_tabs_message=All+links+checked+or+queued.')
        
    security.declareProtected(ManagePortal, 'manage_checkAll')
    def manage_forceCheckAll(self):
        """ZMI method for checking all links."""
        self.forceCheckAllLinks()
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_checker?manage_tabs_message=All+links+forced+into+que.')

    security.declareProtected(ManagePortal, 'manage_findLinks')
    def manage_findLinks(self):
        """ZMI method for finding new links."""
        count = self.findLinks()
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_checklinks?manage_tabs_message=%s+links+added.+%s+links+removed.' % count) 
            
    security.declareProtected(ManagePortal, 'manage_activateWorkerThread')
    def manage_activateWorkerThread(self):
        """ZMI method for activating a new worker thread."""
        try:
            self.activateWorkerThread()
            message = "Activated worker thread."
        except TooManyThreadsRunning:
            message = "Already %s running." % active_threads
            
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checker?manage_tabs_message=%s' % message)

    security.declareProtected(ManagePortal, 'manage_activateSchedulerThread')
    def manage_activateSchedulerThread(self):
        """ZMI method for activating a new scheduler thread."""
        self.activateSchedulerThread()
        message = "Scheduler+activated"
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checker?manage_tabs_message=%s' % message)

    #
    # Interface methods
    #
    def __init__(self):
        self._retrievers = PersistentMapping()

        # register all global retrievers
        for type, retriever in retrievers.retrievers.items():
            self.registerRetriever(type, retriever)

    security.declarePrivate("getLinkId")
    def getLinkId(self, url):
        """The ID is the md5 sum of the link."""
        if url is None:
                import pdb; pdb.set_trace() 
        return md5(url).hexdigest()

    security.declarePrivate("getInfoForURL")
    def getInfoForURL(self, url):
        """Get an info object for this url"""
        id = self.getLinkId(url)
        info = getattr(self, id, None)
        if info is None:
            info = LinkInfo(url)
            self._setObject(id, info)
            info = getattr(self, id)
        return info
        
    def manage_afterAdd(self, container, object):
        # The link catalog that manages links
        self.manage_addProduct['ZCatalog'].manage_addZCatalog('link_catalog', 'Link Catalog')
        self.link_catalog.addIndex('url', 'FieldIndex')
        self.link_catalog.addIndex('state', 'KeywordIndex')
        self.link_catalog.addIndex('objects', 'KeywordIndex')
    
    ##########################################################################
    ## UID methods
    security.declarePrivate('getUIDForObject')
    def getUIDForObject(self, object):
        try:
            return object.portal_linkchecker_uid
        except AttributeError:
            uid_a = object.getId()
            uid_b = ""
            uid = lambda: "%s-%s" % (uid_a, uid_b)
            while uid() in self.portal_catalog.\
                                Indexes['portal_linkchecker_uid'].uniqueValues():
                uid_b = time.time()

            object.portal_linkchecker_uid = uid()
            self.portal_catalog.reindexObject(object)
            return object.portal_linkchecker_uid

    security.declarePrivate('getObjectForUID')
    def getObjectForUID(self, uid):
        results = self.portal_catalog(portal_linkchecker_uid=uid)
        if not results:
                return None
        else:
                return results[0].getObject()

    ## registry functions for link information retrieval
    security.declareProtected(ManagePortal, 'registerRetriever')
    def registerRetriever(self, portal_type, function_ref):
        assert not self._retrievers.has_key(portal_type), \
            "Duplicate portal_type registration (%s)" % portal_type
        self._retrievers[portal_type] = function_ref

    def getRetrievableTypes(self):
        return self._retrievers.keys()

    def getRetriever(self, type):
        return self._retrievers[type]
        
    security.declareProtected(ManagePortal, 'checkLink')
    def checkLink(self, link):

        now = time.time()
        if link.lastcheck:
            age = link.lastcheck.millis() / 1000
        else:
            age = 0
        
        # Cache invalid? Requeue an update and return "unknown"
        if now - age > self.maximumCacheAge:
            self.__queueAddLink(link)
            return "grey"

        # Update required?
        if now - age > self.gracefulUpdateAge:
            self.__queueAddLink(link)
            return "grey"
        
        return link.state

    security.declareProtected(ManagePortal, 'checkAllLinks')
    def checkAllLinks(self):
        for link in self.objectValues(["Link Information"]):
            self.checkLink(link)

    security.declareProtected(ManagePortal, 'checkAllLinks')
    def forceCheckAllLinks(self):
        for link in self.objectValues(["Link Information"]):
            self.__queueAddLink(link)

    security.declarePrivate('clearLinks')
    def clearLinks(self):
        """Removes all links that aren't referenced anymore."""
        to_delete = []
        for x in self.objectValues(["Link Information"]):
            if len(x.objects) == 0:
                to_delete.append(x.getId())
        self.manage_delObjects(to_delete)

    security.declareProtected(ManagePortal, 'findLinks')
    def findLinks(self):
        count = 0
        links = []

        # Clear all uid references
        for x in self.objectValues(["Link Information"]):
            x.clearObjects()

        # gather all objects that are of a type we can check for links
        for key in self.getRetrievableTypes():
            objects = self.portal_catalog(portal_type=key)
            retriever = self.getRetriever(key)
            for ob in objects:
                ob = ob.getObject()     # resolve catalog
                retriever(self, ob)

        # Delete old references
        self.clearLinks()

        return 0, 0

    # XXX this might be wrong
    security.declareProtected('Modify portal content', 'findLinksForObject')
    def findLinksForObject(self, object):
        # Delete old references for this object
        links = self.link_catalog(objects=[self.getUIDForObject(object)])
        for x in links:
            x = x.getObject()
            x.removeObject(object)

        # Update links
        type = object.Type()
        if not type in self.getRetrievableTypes():
            return
        retriever = self.getRetriever(type)
        retriever(self, object)

        # Remove stale links
        self.clearLinks()

    security.declareProtected(ManagePortal, 'showLinks')
    def showLinks(self):
        """Show a grouped list of link data."""
        data = self._cache.values()

        # group them all
        grouped = {}
        for x in data:
            group = x.getGroupId()
            if not grouped.has_key(group):
                grouped[group] = []
            grouped[group].append(x)
        return grouped
        
        return self._cache.values()
                
    security.declareProtected(ManagePortal, 'setMaximumCacheAge')
    def setMaximumCacheAge(self, age):
        self.maximumCacheAge = age

    security.declareProtected(ManagePortal, 'setPeriod')
    def setPeriod(self, period):
        self.checkingPeriod = int(period)

    security.declareProtected(ManagePortal, 'setGracefulUpdateAge')
    def setGracefulUpdateAge(self, age):
        assert age < self.maximumCacheAge, \
            "Graceful update age is larger than maximum cache age."
        self.gracefulUpdateAge = age
     
    security.declareProtected(ManagePortal, 'setDefaultURIPrefix')
    def setDefaultURIPrefix(self, defaultURIPrefix):
        assert type(defaultURIPrefix) in StringTypes
        self.defaultURIPrefix = defaultURIPrefix
        
    security.declareProtected(ManagePortal, 'queueSize')
    def queueSize(self):
        global retrieval_queue
        return retrieval_queue.qsize()

    security.declareProtected(ManagePortal, 'activateWorkerThread')
    def activateWorkerThread(self):
        if not pool_threads.acquire(0):
            raise TooManyThreadsRunning
        global thread_count
        thread_count += 1
        thread = URLCheckerThread(retrieval_queue)
        thread.start()

    security.declareProtected(ManagePortal, 'activateSchedulerThread')
    def activateSchedulerThread(self):
        thread = LinkSchedulerThread(retrieval_queue, self.getPhysicalPath())
        thread.start()

    security.declareProtected(ManagePortal, 'countWorkingThreads')
    def countWorkingThreads(self):
        global thread_count
        return thread_count
        
    ### Private methods
    security.declarePrivate('__queueAddLink')
    def __queueAddLink(self, info):
        """Schedules a new check.

        Info must be a ILinkInfo object."""
        zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "Queue", "\tManually queueing %s\n" % info.url)
        retrieval_queue.put((self.getPhysicalPath(), info.getId())) 
        info.queued = 1

    ### UI convenience methods
    security.declarePublic("isManaged")
    def isManaged(self, object):
        """Returns true/false for objects if they can be checked."""
        try:
           type = object.Type()
        except:
            return 0

        return type in self.getRetrievableTypes()

    security.declareProtected("Modify portal content", "getLinksForObject")
    def getLinksForObject(self, object):
        """Return a list of link information objects that are valid for this object."""
        links = self.link_catalog(objects=[self.getUIDForObject(object)])
        links = [ x.getObject() for x in links ]
        return links

    security.declareProtected("View", "getLinksForAuthenticatedMember")
    def getGroupedLinksForAuthenticatedMember(self):
        """Returns a list of link information objects that are valid for the authenticated user."""
        sm = getSecurityManager()
        user = sm.getUser()
        links = self.portal_catalog(Creator=user.getId())
        uid = [ self.getUIDForObject(x.getObject()) for x in links ]
        links = self.link_catalog(uid=uid)
        links = [ x.getObject() for x in links ]
        links = [ x for x in links if x is not None ]
        groups = {'red':[], 'orange':[], 'grey':[], 'green':[]}
        for link in links:
                groups[link.state].append(link)
                        
        return groups

    security.declareProtected("Manage portal", "getBalancedScoreCard")
    def getBalancedScoreCard(self):
        status = {}

        # Total amount of links (distinct)
        status['total_distinct'] = len(self.objectValues(["Link Information"]))

        # Total amount of objects
        status['total_documents'] = len(self.link_catalog.Indexes['objects'].uniqueValues())

        # Documents per link
        values = []
        for x in self.objectValues(["Link Information"]):
            values.append(len(x.objects))
        status['documents_per_link_min'] = min(values)
        status['documents_per_link_max'] = max(values)
        status['documents_per_link_avg'] = "%.02f" % reduce(lambda x,y:(x+y)/2.0, values)

        # Total amount of links
        status['total'] = reduce(lambda x,y:x+y, values)
        
        # Links per document
        documents = {}
        for x in self.objectValues(["Link Information"]):
            for ob in x.objects:
                count = documents.get(ob, 0)
                count += 1
                documents[ob] = count
        documents = documents.values()
        status['links_per_document_min'] = min(documents)
        status['links_per_document_max'] = max(documents)
        status['links_per_document_avg'] = "%.02f" % reduce(lambda x,y:(x+y)/2.0, documents)

        # Links per state (total)
        states = {'red':0, 'grey':0, 'green':0, 'orange':0}
        for x in self.objectValues(["Link Information"]):
            states[x.state] += 1
        for x in states.keys():
            status['total_%s_links' % x] = states[x]
            status['relative_%s_links' % x] = "%0.2f" % (float(states[x]) / status['total_distinct'] * 100)

        # Documents with state ...
        states = {'red': [], 'grey': [], 'green': [], 'orange': []}
        for x in self.objectValues(["Link Information"]):
            # We need to make it this complicated because we don't want to count documents
            # multiple times
            for y in x.objects:
                if y not in states[x.state]:
                    states[x.state].append(y)
        for x in states.keys():
            status['total_%s_documents' % x] = len(states[x])
            status['relative_%s_documents' % x] = "%.02f" % (float(len(states[x])) / status['total_documents'] * 100)

        return status
        
InitializeClass(LinkCheckerTool)

#### Threading code.
# We only use one worker thread. Object references
# should be stable to work with.

class LinkSchedulerThread(Thread):

    def __init__(self, queue, linkchecker):
        Thread.__init__(self, name="LinkScheduler")
        self.queue = queue
        self.linkchecker = linkchecker

    def run(self):
        zLOG.LOG('CMFLinkChecker', 0, 'Initialized', '\tScheduling thread has been initialized.\n')

        # calculate timing for scheduler
        check_app = None
        links = []
        while 1:
            if check_app:
                    get_transaction().abort()   # We don't do write-requests anyway
                    check_app._p_jar.sync()
                    check_app._p_jar.close()
                    check_app = None
            DB=Globals.DB
            app=ZODB.ZApplication.ZApplicationWrapper(
                DB, 'Application', OFS.Application.Application, (),
                Globals.VersionNameName)
            check_app=app()
            
            o = check_app
            # find the linkchecker object
            for p in self.linkchecker:
                if p:
                    o=getattr(o,p)
            linkchecker = o
            if not links:
                # try to get a zodb connection
                
                period = linkchecker.checkingPeriod
                links = linkchecker.objectIds(["Link Information"])
                shuffle(links)
                pause = period/float(len(links))
                zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "Automatic Scheduler", "\tPause set to %s seconds\n" % pause)

            link = links.pop()
            link = getattr(linkchecker, link)
            zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "Automatic Scheduler", "\tQueuing %s\n" % link.url)
            self.queue.put((linkchecker.getPhysicalPath(), link.getId()))

            time.sleep(pause)
            
class URLCheckerThread(Thread):

    def __init__(self, queue):
        Thread.__init__(self, name="URLChecker")
        self.queue = queue

    def run(self):
        zLOG.LOG('CMFLinkChecker', 0, 'Initialized', '\tWorker thread has been initialized.\n')
        # try to get a zodb connection
        check_app = None
        while 1:
            if check_app:
                get_transaction().commit()
                check_app._p_jar.sync()
                check_app._p_jar.close()
                check_app = None

            # we are doing cooperative multitasking
            # this thread doesn't need too much power
            time.sleep(0.1)
            try:
                linkchecker, key = self.queue.get(block=0)
            except Queue.Empty:
                time.sleep(1)
                continue
            
            # We got work, so setup the database connection.
            DB=Globals.DB
            app=ZODB.ZApplication.ZApplicationWrapper(
                DB, 'Application', OFS.Application.Application, (),
                Globals.VersionNameName)
            check_app=app()
            
            o = check_app
            # find the linkchecker object
            for p in linkchecker:
                if p:
                    o=getattr(o,p)
            linkchecker = o

            try:
                item = getattr(linkchecker, key)
            except AttributeError:
                continue
            zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "Checker", "\tChecking %s\n" % item.url)

            # Links need to revalidate everytime otherwise they are in alert state.
            item.queued = 0
            item.lastcheck = DateTime()

            # only check http links. other links are ... hmm
            # ... unknown ...
            type, url = urllib.splittype(item.url)
            if not type == 'http':
                item.state = 'grey'
                item.reason = "Protocol not supported by CMFLinkChecker"
                item.lasttime = 0
                item.headers = {}
                continue
            
            # get a connection to the server
            host, path = urllib.splithost(url)
            if host is None:
                item.state = "red"
                item.reason = "Invalid Link: No hostname defined"
                continue
            host, port = urllib.splitnport(host, 80)
            
            conn = httpclient.myHTTPConnection(host, port, 2)
            starttime = time.time()
            # request the link
            try:
                conn.request('GET', path)
            except:
                import sys
                item.state = "red"
                item.reason = sys.exc_value[1]          # this is http specific
                continue

            # look into the response
            response = None
            while not response:
                try:
                    now = time.time()
                    response = conn.getresponse()
                except ResponseNotReady:
                    if starttime - now > 4:
                        break
                except IOError:
                    break
                time.sleep(0.2)

            conn.close()
            if response is not None:
                response.close()

            item.lasttime = now - starttime
            item.headers = {}

            # what if we didn't got a response?
            if not response:
                item.state = 'red'
                item.reason = 'unknown'
                continue

            if not http_codes.has_key(response.status):
                item.state = 'grey'
                item.reason = 'unknown HTTP response'
                continue
            item.state, item.reason = http_codes.get(response.status)
                
        zLOG.LOG('CMFLinkChecker', 0, 'Stopped', '\tWorker thread stopped.\n')
        global thread_count
        thread_count -= 1

