##############################################################################
#
# Copyright (c) 2003 gocept gmbh & co. kg. All rights reserved.
#
# See also LICENSE.txt
#
##############################################################################
"""CMF link checker tool - link information object

$Id: LinkInfo.py,v 1.5 2003/12/17 19:18:27 ctheune Exp $"""

from Globals import Persistent
from Acquisition import Implicit
from OFS.SimpleItem import SimpleItem
from interfaces import ILinkInformation
from urllib import urlencode

from Products.CMFCore.utils import getToolByName
from httpcodes import http_codes


class LinkInfo(SimpleItem):

    meta_type = "Link Information"

    __implements__ = (ILinkInformation, )
    __allow_access_to_unprotected_subobjects__ = 1      # XXX make accessor functions

    owner = None
    objects = []
    url = None
    lastcheck = None
    lasttime = None
    queued = 0
    state = "grey"
    headers = {}
    reason = ""

    def __init__(self, url):
        self.objects = []
        self.setUrl(url)

    def _p_resolveConflict(self, oldState, savedState, newState):
        return newState

    def getId(self):
        return self.aq_parent.getLinkId(self.url)

    def title(self):
        return self.url
        
    def addObject(self, object):
        """Add this object to the object list."""
        uid = self.portal_linkchecker.getUIDForObject(object)
        if not uid in self.objects:
            self.objects.append(uid)
            self._p_changed = 1
            self.index()
        
    def removeObject(self, object):
        """Remove this uid from the object list."""
        uid = self.portal_linkchecker.getUIDForObject(object)
        if uid in self.objects:
            self.objects.remove(uid)
            self._p_changed = 1
            self.index()

    def clearObjects(self):
        """Remove all uid references from the object list."""
        self.objects = []
        self._p_changed = 1
        self.index()

    def getObjects(self):
        """Return a reference to the object."""
        objects = [ self.portal_linkchecker.getObjectForUID(object) 
                    for object in self.objects ]
        return objects

    def setUrl(self, url, REQUEST=None):
        """Set the URL.

        Sets the URL to the given value, trying to make relative paths absolute.
        (Requires a REQUEST object)"""
        self.url = url

        if REQUEST is None:
            return

        # Try to normalize relative links
        oldurl = self.url
        if self.url.startswith("/"):
            # Maybe we have a default prefix on the tool?
            lt = getToolByName(self, 'portal_linkchecker', None)
            self.url = getattr(lt, 'defaultURIPrefix', REQUEST.BASE0) + self.url
        elif not ':' in self.url:
            try:
                self.url = REQUEST.BASE0+'/'.join(self.object)+'/'+self.url
            except:
                pass

        self.index()

    def getGroupId(self):
        return self.state

    # Catalog support

    def manage_afterAdd(self, container, object):
        self.index()

    def manage_beforeDelete(self, item, container):
        self.unindex()

    def index(self):
        self.link_catalog.catalog_object(self, '/'.join(self.getPhysicalPath()))

    def unindex(self):
        self.link_catalog.uncatalog_object('/'.join(self.getPhysicalPath()))
