##############################################################################
#
# Copyright (c) 2003 gocept gmbh & co. kg. All rights reserved.
#
# See also LICENSE.txt
#
##############################################################################
"""CMF link checker tool - link retriever functions

$Id: retrievers.py,v 1.5 2003/12/17 19:18:27 ctheune Exp $"""

from Products.CMFDefault.Link import Link
from Products.CMFDefault.Document import Document
from Products.CMFCalendar.Event import Event

from StructuredText import Basic
from StructuredText.DocumentClass import StructuredTextLink, DocumentClass

from LinkInfo import LinkInfo

import lchtmllib
from sgmllib import SGMLParseError
import formatter

def retrieverLink(linkchecker, object):
    """Usable for CMFDefault.Link objects."""
    assert isinstance(object, Link)
    url = object.getRemoteUrl()
    info = linkchecker.getInfoForURL(url)
    info.addObject(object)
    
def retrieverEvent(linkchecker, object):
    """Usable for CMFDefault.Event objects."""
    assert isinstance(object, Event)
    url = object.event_url
    info = linkchecker.getInfoForURL(url)
    info.addObject(object)
    
def retrieveDocument(linkchecker, object):
    """Usable for CMFDefault.Document objects (in HTML and STX mode)"""
    assert isinstance(object, Document)
    
    # Handle document modes
    if object.text_format == 'structured-text':
        links = retrieveSTX(object.text)
    elif object.text_format == 'html':
        links = retrieveHTML(object.text)
    else:
        links = []

    # Handle link registration
    for link in links:
        info = linkchecker.getInfoForURL(link)
        info.addObject(object)

def retrieveSTX(text):
    """Retrieve links from STX data."""
    data = Basic(text)
    data = DocumentClass()(data)
    
    stack = data.getChildren()
    links = []
    while stack:
        element = stack.pop()
        if isinstance(element, StructuredTextLink):
            # Handle images seperately
            element = element.href
            if element.startswith('img:'):
                element = element[4:]
            links.append(element)
        try:
            stack.extend(element.getChildren())
        except AttributeError:
            pass

    return links

def retrieveHTML(text):
    parser = lchtmllib.LCHTMLParser(formatter.NullFormatter())
    try:
        parser.feed(text)
    except SGMLParseError:
        # SGMLLib seems to die on bad HTML sometimes. (At least with python2.1)
        pass
    parser.close()
    return tuple(parser.anchorlist) + tuple(parser.imglist)

# simple registries

retrievers = { 
      'Link'     :   retrieverLink,
      'Event'    :   retrieverEvent,
      'Favorite' :   retrieverLink,
      'Document' :   retrieveDocument,
      'News Item' :  retrieveDocument }
